// RS232.CPP
// /-------------------------------------------------------------------------\
// |      Klasse zum Ansprechen der RS232-Schnittstelle unter DOS            |
// |                                                                         |
// |                      (C) 30.3.1995 Jrgen Hasch                         |
// \-------------------------------------------------------------------------/
// V1.1a / Borland-C++ 3.1
// !MINIMALVERSION!
//  -> nur linearer Empfangspuffer / kein Sendepuffer
//

#include <dos.h>
#include <stdarg.h>
#include <stdio.h>
#include <iostream.h>

static int base;                // Portadresse
volatile int inptr;             // Zhler Empfangspuffer
static int rxptr;
static char *inbuf;             // Zeiger auf Empfangspuffer

#include "rs232.hpp"
//
// Interrupt-Handler fr UART 16550 (FIFO)
//
void far interrupt RS232::fifo_irq(...)
{
register c;
  asm  cli
  asm  mov  DX,word ptr base            // Portadresse
  asm  add  DX,IIR                      // interrupt identification register
  asm  in   AL,DX
  asm  add  DX,LSR-IIR
  asm  and  AL,07H
  asm  cmp  AL,04                       // Interruptquelle: Zeichen empfangen ?
  asm  je   getch
  asm  cmp  AL,0CH                      // noch Zeichen im FIFO
  asm  jne  endint
getch:
  asm  sub  DX,LSR
  asm  in   AL,DX                       // Zeichen aus Register lesen uns speichern
  asm  mov  byte ptr c,AL
  inbuf[inptr] = c; inptr++;
  asm  add  DX,LSR
  asm  in   AL,DX
  asm  AND  AL,01
  asm  jne  getch                       // kompletten FIFO auslesen
endint:
  asm  mov  DX,ICR
  asm  mov  AL,EOI
  asm  out  DX,AL
  asm sti
  return;
}

//
// Interrupt-Handler fr UART 8250
//
void far interrupt RS232::nofifo_irq(...)
{
register c;
  asm  cli
  asm  mov  DX,word ptr base            // Portadresse
  asm  add  DX,IIR                      // interrupt identification register
  asm  in   AL,DX
  asm  and  AL,07H
  asm  cmp  AL,04                       // Interruptquelle: Zeichen empfangen ?
  asm  jne  endint

  asm  sub  DX,IIR
  asm  in   AL,DX                       // Zeichen aus Register lesen uns speichern
  asm  mov  byte ptr c,AL
  inbuf[inptr] = c; inptr++;
endint:
  asm  mov  DX,ICR
  asm  mov  AL,EOI
  asm  out  DX,AL
  asm sti
  return;
}

//
// Interrupt-Vektor setzen
//
void RS232::my_setvect(unsigned char numm, void far interrupt (RS232::*func)(...))
{
asm     mov     ah, 025h
asm     mov     al, numm
asm     push    ds
asm     lds     dx, dword ptr func
asm     int     021h
asm     pop     ds
}

//
// RS232-Schnittstelle initialisieren
//
void RS232::RS232Init()
{
  inbuf = (char *)malloc(IBUF_LEN);  // Empfangspuffer

  OldVect = getvect(irq+MCI);             // alten Interrupt-Vektor sichern

// alt
  disable();                              // Interrupts ausschalten
  outportb(base+MCR, DTR|RTS);            // RTS und DTR einschalten
  int temp = inportb(base+LSR);           // read serial port line status
  temp = inportb(base+RXR);               // read char
  temp = inportb(IMR);                    // get interrupt settings
  temp = temp & ((IER<<irq)^0xff);        // turn on serial interrupt
  outportb(IMR, temp);                    // send out new int settings
  outportb(base+IER, 0x01);               // enable RX interrupt
  outportb(base+MCR, inportb(base+MCR) | 0x0a);

  // UART 8250
  fifoen = 0;
  my_setvect(irq+MCI,&RS232::nofifo_irq);
  enable();

  inptr=0;                                // Puffer leeren
  rxptr=0;
}

RS232::RS232(unsigned char p,unsigned long b)
{
  baud = b;
  base = COM_BASE[p];
  irq  = COM_IRQ[p];
  RS232Init();
  setBaud(baud);
}

RS232::RS232(unsigned long b, int portaddr, int portint)
{
  baud = b;
  base = portaddr;
  irq = portint;
  RS232Init();
  setBaud(baud);
}

//
// Destruktor
//
RS232::~RS232()
{
  disable();
  int temp = inportb(IMR) | (1 << irq);  // HW-Interrupt ausschalten
  outportb(IMR, temp);
  outportb(base+IER, 0x00);              // UART-Interrupts sperren
  outportb(base+MCR, 0);                 // RTS und DTR lschen
  enable();
  setvect(irq+MCI,OldVect);              // alten Handler wieder einsetzen
  free(inbuf);
}

//
// Baudrate setzen
//
void RS232::setBaud(unsigned long b)
{
    if (b > 49 && b < 115201L)
    {
	float rate = 115200.0 / ((float)b);
	baud = (unsigned int)rate;

	outportb(base+LCR, inportb(base+LCR) | 0x80);
	outportb(base, (baud & 0x00ff));
	outportb(base+IER, ((baud >> MCI) & 0x00ff));
	outportb(base+LCR, inportb(base+LCR) & 0x7f);
    }
}

//
// Zeichen senden
//
void RS232::dotx(char ch)
{
  disable();
  while ( !(inportb(base+LSR) & 0x20))
    ;
  outportb(base+TXR,ch);
  enable();
}

//
// Zeichen aus Empfangspuffer holen
//
char RS232::rxchar()
{
 return dorx();
}

//
// Wort aus Empfangspuffer holen
//
WORD RS232::rxword()
{
WORD wrd;
  wrd = (unsigned)rxchar();
 wrd += 256*rxchar();
 return wrd;
}

//
// Zeichen senden
//
void RS232::txchar(char ch)
{
  dotx(ch);
}

//
// Wort senden
//
void RS232::txword(WORD wrd)
{
  dotx((char) (wrd >> 8)&255);
  dotx(wrd & 255);
}

//
// Zeichen aus Empfangspuffer lesen
//
char RS232::dorx(void)
{
char ch=0;
int cnt;
  cnt = 1;
  while( rxptr >= inptr && cnt < 100)  // wait until char received or timeout
    { delay(1); cnt++; }
 
  if (inptr >= rxptr)
    ch = inbuf[rxptr++];
  return ch;
}

//
// String senden
//
void RS232::txstring(char *str)
{
    enable();
    while (*str)
    {
	if (*str=='\n')
	    txchar('\r'), txchar('\n');
	else
	    txchar(*str);
	str++;
    }
}

//
// formatierten String senden
//
void RS232::printf(char *format, ...)
{
    va_list argptr;
    char str[256];

    va_start(argptr,format);            // access argument list
    vsprintf(str,format,argptr);        // create string using argument list
    va_end(argptr);                     // end access of argument list

    txstring(str);                      // output the formatted string
}
//
// Empfangspuffer lschen
//
void RS232::flushInBuf(void)
{
  disable();
  inptr=0; rxptr = 0;
  enable();
}
//
// Status der DTR-Leitung abfragen
//
int RS232::dtr(void)
{
  return inportb(base + MCR) & 1;
}
//
// DTR setzen
//
void RS232::setDtr(void)        // DTR setzen
{
  outportb(base+MCR, inportb(base+MCR) |0x01);
}
//
// DTR lschen
//
void RS232::dropDtr(void)              // DTR lschen
{
  outportb(base+MCR, inportb(base+MCR) & 0x00fe);
}
